@@ -211,7 +211,22 @@ class OpenSSLGenerator(ConfigGenerator):
         'RSA3072-SPHINCSSHAKE128FSIMPLE': '?rsa3072_sphincsshake128fsimple',
     }
 
-    group_map = {
+    group_pq_map = {
+        # provider-only, so, optional (openssl#23050) + marked experimental
+        'MLKEM512': '?mlkem512',
+        'P256-MLKEM512': '?p256_mlkem512',
+        'X25519-MLKEM512': '?x25519_mlkem512',
+        'MLKEM768': '?mlkem768',
+        'P384-MLKEM768': '?p384_mlkem768',
+        'X448-MLKEM768': '?x448_mlkem768',
+        'MLKEM768-X25519': '?X25519MLKEM768:?x25519_mlkem768',  # new/old name
+        'P256-MLKEM768': '?SecP256r1MLKEM768:?p256_mlkem768',  # new/old name
+        'MLKEM1024': '?mlkem1024',
+        'P521-MLKEM1024': '?p521_mlkem1024',
+        'P384-MLKEM1024': '?SecP384r1MLKEM1024:?p384_mlkem1024',  # new/old
+    }
+
+    group_classic_map = {
         'SECP224R1': 'secp224r1',
         'SECP256R1': 'secp256r1',
         'SECP384R1': 'secp384r1',
@@ -226,21 +241,6 @@ class OpenSSLGenerator(ConfigGenerator):
         'BRAINPOOL-P256R1': 'brainpoolP256r1:?brainpoolP256r1tls13',
         'BRAINPOOL-P384R1': 'brainpoolP384r1:?brainpoolP384r1tls13',
         'BRAINPOOL-P512R1': 'brainpoolP512r1:?brainpoolP512r1tls13',
-        # provider-only, so, optional (openssl#23050) + marked experimental
-        'KYBER768': '?kyber768',
-        'X25519-KYBER768': '?x25519_kyber768',
-        'P256-KYBER768': '?p256_kyber768',
-        'MLKEM512': '?mlkem512',
-        'P256-MLKEM512': '?p256_mlkem512',
-        'X25519-MLKEM512': '?x25519_mlkem512',
-        'MLKEM768': '?mlkem768',
-        'P384-MLKEM768': '?p384_mlkem768',
-        'X448-MLKEM768': '?x448_mlkem768',
-        'X25519-MLKEM768': '?X25519MLKEM768:?x25519_mlkem768',  # new/old name
-        'P256-MLKEM768': '?SecP256r1MLKEM768:?p256_mlkem768',  # new/old name
-        'MLKEM1024': '?mlkem1024',
-        'P521-MLKEM1024': '?p521_mlkem1024',
-        'P384-MLKEM1024': '?SecP384r1MLKEM1024:?p384_mlkem1024',  # new/old
     }
 
     @classmethod
@@ -335,8 +335,20 @@ class OpenSSLGenerator(ConfigGenerator):
         sig_algs = [cls.sign_map[i] for i in p['sign'] if i in cls.sign_map]
         s += 'SignatureAlgorithms = ' + ':'.join(sig_algs) + '\n'
 
-        groups = [cls.group_map[i] for i in p['group'] if i in cls.group_map]
-        s += 'Groups = ' + ':'.join(groups) + '\n'
+        # Separate groups into PQ and classic groups, generate them as follows:
+        # `*first_pq:rest_pq/*first_classic:rest_classic`.
+        # This way servers will prefer any PQ over any classic,
+        # and clients will send key_shares for top priority PQ
+        # and top priority classic groups
+        groups_pq = [cls.group_pq_map[i] for i in p['group']
+                     if i in cls.group_pq_map]
+        groups_classic = [cls.group_classic_map[i] for i in p['group']
+                          if i in cls.group_classic_map]
+        group_classes = (
+            (['*' + ':'.join(groups_pq)] if groups_pq else [])
+            + (['*' + ':'.join(groups_classic)] if groups_classic else [])
+        )
+        s += 'Groups = ' + '/'.join(group_classes) + '\n'
 
         if policy.enums['__ems'] == 'RELAX':
             s += 'Options = RHNoEnforceEMSinFIPS\n'
